import json
import os

from dapla_suv_tools._internals.integration.api_client import SuvApiClient
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util import constants
from dapla_suv_tools._internals.util.validators import instance_str_validator
from dapla_suv_tools._internals.util.decorators import result_to_dict


END_USER_API_BASE_URL = os.getenv("SUV_END_USER_API_URL", "")

client = SuvApiClient(base_url=END_USER_API_BASE_URL)


@result_to_dict
@SuvOperationContext(validator=instance_str_validator)
def get_instance(
    self, *, instance_id: str, context: SuvOperationContext
) -> OperationResult:
    """
    Gets an 'instance' based on it's owner id and guid.
    :param instance_id: The combined instance's owner id and guid separated by a '/' (e.g. '123451213/12345678-1234-1234-1234-123456789012').
    :param context: Operation context.  This is injected by the underlying pipeline.  Adding a custom context will result in an error.
    :return: a json object containing instance data.
    """
    try:
        content: str = client.get(
            path=f"{constants.INSTANCE_PATH}/{instance_id}", context=context
        )
        content_json = json.loads(content)
        context.log(message="Fetched 'instance' with id '{instance_id}'")
        return OperationResult(value=content_json, log=context.logs())

    except Exception as e:
        context.set_error(f"Failed to fetch for id {instance_id}", e)
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )



from datetime import date
import json
import os
from typing import Optional
from dapla_suv_tools.pagination import PaginationInfo


from ssb_altinn3_util.models.skjemadata.skjemadata_request_models import (
    PeriodeRequestModel,
)

from dapla_suv_tools._internals.integration.api_client import SuvApiClient
from dapla_suv_tools._internals.integration import user_tools
from dapla_suv_tools._internals.util import constants
from dapla_suv_tools._internals.util.decorators import result_to_dict
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util.validators import (
    periode_id_validator,
    skjema_id_validator,
)


END_USER_API_BASE_URL = os.getenv("SUV_END_USER_API_URL", "")

client = SuvApiClient(base_url=END_USER_API_BASE_URL)


def bool_to_str(value: bool | None, field_name: str = "") -> str | None:
    """Convert a boolean value to a string representation."""
    if field_name == "har_skjemadata":
        return "Y" if value is True else "N"
    return "1" if value is True else "N" if value is False else None


def convert_to_bool(value: str) -> bool:
    """Convert '1' or 'Y' to True, 'N' to False, otherwise return the original value."""
    return True if value in ['1', 'Y'] else False if value == 'N' else value



def _get_non_paged_result(
    path: str, max_results: int, filters: str, context: SuvOperationContext
) -> str:
    if max_results > 0:
        return client.post(
            path=f"{path}?size={max_results}&order_by=versjon&asc=false",
            body_json=filters,
            context=context,
        )

    items = []
    total = 1
    page = 1

    while len(items) < total:
        response = client.post(
            path=f"{path}?page={page}&size=100&order_by=versjon&asc=false",
            body_json=filters,
            context=context,
        )

        response_json = json.loads(response)
        total = int(response_json["total"])
        items.extend(response_json["results"])
        page += 1

    return json.dumps({"results": items})


def _get_paged_result(
    path: str, paging: PaginationInfo, filters: str, context: SuvOperationContext
) -> str:
    return client.post(
        path=f"{path}?page={paging.page}&size={paging.size}&order_by=versjon&asc=false",
        body_json=filters,
        context=context,
    )


@result_to_dict
@SuvOperationContext(validator=periode_id_validator)
def get_periode_by_id(
    self, *, periode_id: int, context: SuvOperationContext
) -> OperationResult:
    """
    Retrieves a period by its ID.

    Parameters:
    ------------
    periode_id: int
        The ID of the period to retrieve.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.  Adding a custom context will result in an error

    Returns:
    --------
    OperationResult:
        An object containing the period information if found, or an error message if the retrieval fails.

    Example:
    ---------
    result = get_periode_by_id(periode_id=123)

    """
    try:
        content: str = client.get(
            path=f"{constants.PERIODE_PATH}/{periode_id}", context=context
        )
        content_json = json.loads(content)

        # Apply conversion to the relevant fields
        content_json['vis_oppgavebyrde'] = convert_to_bool(content_json.get('vis_oppgavebyrde', ''))
        content_json['vis_brukeropplevelse'] = convert_to_bool(content_json.get('vis_brukeropplevelse', ''))
        content_json['har_skjemadata'] = convert_to_bool(content_json.get('har_skjemadata', ''))

    
        context.log(message=f"Fetched periode with periode_id '{periode_id}'")

        return OperationResult(value=content_json, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to fetch for id {periode_id}", e)

        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=skjema_id_validator)
def get_perioder_by_skjema_id(
    self,
    *,
    skjema_id: int,
    periode_type: Optional[str] = None,  # Optional type parameter
    periode_nr: Optional[int] = None,  # Optional nummer parameter
    periode_aar: Optional[int] = None,
    delreg_nr: Optional[int] = None,
    enhet_type: Optional[int] = None,
    max_results: int = 0,
    latest_only: bool = False,
    pagination_info: Optional[PaginationInfo] = None,
    context: SuvOperationContext,
) -> OperationResult:
    """
     Retrieves periods associated with a specific schema ID.

    Parameters:
    ------------
    skjema_id: int
        The skjema_id of the schema for which to retrieve periods.
    periode_type: Optional[str]
        The type of the period to filter by. If None, periods of any type will be retrieved.
    periode_nr: Optional[int]
        The number of the period to filter by. If None, periods of any number will be retrieved.
    periode_aar: Optional[int]
        The year of the period to filter by. If None, periods of any year will be retrieved.
    delreg_nr: Optional[int]
        The delreg_nr of the period to filter by. If None, periods of any delreg_nr will be retrieved.
    enhet_type: Optional[int]
        The enhet_type of the period to filter by. If None, periods of any enhet_type will be retrieved.
    max_results: int
        Maximum number of results in the result set. A value of 0 will get ALL results. Defaults to 0
    latest_only: bool
        A boolean flag to trigger a special condition. A True value will retrieve the latest periode added. Defaults to False.
    pagination_info: Optional[int]
        An object holding pagination metadata. Defaults to None.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.

    Returns:
    --------
    OperationResult:
        An object containing a list of periods associated with the schema ID, or an error message if the retrieval fails.

    Example:
    ---------
    result = get_perioder_by_skjema_id(
        skjema_id=456, periode_type="KVRT", periode_aar=2023
    )
    """

    try:
        filters = {
            "skjema_id": skjema_id,
            "periode_type": periode_type,
            "periode_nr": periode_nr,
            "periode_aar": periode_aar,
            "delreg_nr": delreg_nr,
            "enhet_type": enhet_type,
        }
        filters_json = json.dumps(filters)
        content: str
        if pagination_info is None:
            content = _get_non_paged_result(
                path="/periode/periode-paged",
                max_results=max_results,
                filters=filters_json,
                context=context,
            )
        else:
            content = _get_paged_result(
                path="/periode/periode-paged",
                paging=pagination_info,
                filters=filters_json,
                context=context,
            )

        result: dict = json.loads(content)

        for item in result["results"]:
            item["vis_oppgavebyrde"] = convert_to_bool(item.get("vis_oppgavebyrde", ''))
            item["vis_brukeropplevelse"] = convert_to_bool(item.get("vis_brukeropplevelse", ''))
            item["har_skjemadata"] = convert_to_bool(item.get("har_skjemadata", ''))

        

        if latest_only:
            context.log(
                message=f"Fetched latest version of 'periode' with skjemda id '{skjema_id}'"
            )
            return OperationResult(value=result["results"][0], log=context.logs())

        context.log(message="Fetched all 'perioder' with skjema id '{skjema_id}'")
        return OperationResult(value=result["results"], log=context.logs())

    except Exception as e:
        context.set_error(f"Failed to fetch periode for skjema id '{skjema_id}'.", e)
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=periode_id_validator)
def update_periode_by_id(
    self,
    *,
    periode_id: int,
    periode_dato: date | None = None,
    delreg_nr: int | None = None,
    enhet_type: str | None = None,
    vis_oppgavebyrde: bool | None = None,
    vis_brukeropplevelse: bool | None = None,
    har_skjemadata: bool | None = None,
    journalnummer: str | None = None,
    context: SuvOperationContext,
) -> OperationResult:
    """
    Updates an existing period with the specified periode_id.

    Parameters:
    ------------
    periode_id: int
        The ID of the period to update.
    periode_dato: Optional[date]
        Date for the period. If None, the existing date will be used.
    delreg_nr: Optional[int]
        delreg_nr. If None, the existing number will be used.
    enhet_type: Optional[str]
        enhet_type. If None, the existing type will be used.
    vis_oppgavebyrde: Optional[bool]
        A boolean flag to indicate visibility of "oppgavebyrde". Defaults to None.
    vis_brukeropplevelse: Optional[bool]
        A boolean flag to indicate visibility of "brukeropplevelse". Defaults to None.
    har_skjemadata: Optional[bool]
        A boolean flag to indicate the presence of schema data. Defaults to None.
    journalnummer: Optional[str]
        Journal number. If None, the existing number will be used.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.

    Returns:
    --------
    OperationResult:
        An object containing the updated period information, or an error message if the update fails.

    Example:
    ---------
    result = update_periode_by_id(
        periode_id=456, vis_oppgavebyrde=True, date(2024,12,15)
    )
    """

    user = user_tools.get_current_user(context)

    periode = get_periode_by_id(self=self, periode_id=periode_id)

    # Convert boolean to string using the helper function.
    vis_oppgavebyrde_str = bool_to_str(vis_oppgavebyrde) if vis_oppgavebyrde is not None else bool_to_str(periode["vis_oppgavebyrde"])
    vis_brukeropplevelse_str = bool_to_str(vis_brukeropplevelse) if vis_brukeropplevelse is not None else bool_to_str(periode["vis_brukeropplevelse"])
    har_skjemadata_str = bool_to_str(har_skjemadata, "har_skjemadata") if har_skjemadata is not None else bool_to_str(periode["har_skjemadata"], "har_skjemadata")

    

    body = {
        "id": periode_id,
        "skjema_id": periode["skjema_id"],
        "periode_type": periode["periode_type"],
        "periode_nr": periode["periode_nr"],
        "periode_aar": periode["periode_aar"],
        "periode_dato": periode_dato.isoformat()
        if periode_dato is not None
        else periode["periode_dato"],
        "delreg_nr": delreg_nr if delreg_nr is not None else periode["delreg_nr"],
        "enhet_type": enhet_type if enhet_type is not None else periode["enhet_type"],
        "vis_oppgavebyrde": vis_oppgavebyrde_str,
        "vis_brukeropplevelse": vis_brukeropplevelse_str,
        "har_skjemadata": har_skjemadata_str,
        "journalnummer": journalnummer
        if journalnummer is not None
        else periode["journalnummer"],
        "endret_av": user,
    }

    try:
        body_json = json.dumps(body)
        content: str = client.put(
            path=f"{constants.PERIODE_PATH}/{periode_id}",
            body_json=body_json,
            context=context,
        )

        result: dict = json.loads(content)


        for key in ['vis_brukeropplevelse', 'vis_oppgavebyrde', 'har_skjemadata']:
            if key in result:
                result[key] = convert_to_bool(result[key])
        
        return OperationResult(value=result, log=context.logs())
    except Exception as e:
        context.set_error(
            f"Failed to update periode for periode_id {periode_id}",
            e,
        )
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=skjema_id_validator)
def update_periode_by_skjema_id(
    self,
    *,
    skjema_id: int,
    periode_type: str,
    periode_aar: int,
    periode_nr: int,
    periode_dato: date | None = None,
    delreg_nr: int | None = None,
    enhet_type: str | None = None,
    vis_oppgavebyrde: bool | None = None,
    vis_brukeropplevelse: bool | None = None,
    har_skjemadata: bool | None = None,
    journalnummer: str | None = None,
    context: SuvOperationContext,
) -> OperationResult:
    """

    Updates an existing period with the specified skjema_id.

    Parameters:
    ------------
    skjema_id: int
        The skjema_id of the period to update.
    periode_type: str
        Periode type of the period to update.
    periode_aar: int
        Year of the period to update.
    periode_nr: int
        Periode number of the period to update.
    periode_dato: Optional[date]
        Date for the period. If None, the existing date will be used.
    delreg_nr: Optional[int]
        delreg_nr. If None, the existing number will be used.
    enhet_type: Optional[str]
        Unit type. If None, the existing type will be used.
    vis_oppgavebyrde: Optional[bool]
        A boolean flag to indicate visibility of "oppgavebyrde". Defaults to None.
    vis_brukeropplevelse: Optional[bool]
        A boolean flag to indicate visibility of "brukeropplevelse". Defaults to None.
    har_skjemadata: Optional[bool]
        A boolean flag to indicate the presence of schema data. Defaults to None.
    journalnummer: Optional[str]
        Journal number. If None, the existing number will be used.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.

    Returns:
    --------
    OperationResult:
        An object containing the updated period information, or an error message if the update fails.

    Example:
    ---------
    result = update_periode_by_skjema_id(
        skjema_id=456, periode_type="KVRT", periode_aar=2023, date(2024,12,15), periode_nr=1
    )
    """

    user = user_tools.get_current_user(context)

    periode_list = get_perioder_by_skjema_id(
        self=self,
        skjema_id=skjema_id,
        periode_type=periode_type,
        periode_aar=periode_aar,
        periode_nr=periode_nr,
    )

    periode = periode_list[0]
 
      # Convert boolean to string using the helper function.
    vis_oppgavebyrde_str = bool_to_str(vis_oppgavebyrde) if vis_oppgavebyrde is not None else bool_to_str(periode["vis_oppgavebyrde"])
    vis_brukeropplevelse_str = bool_to_str(vis_brukeropplevelse) if vis_brukeropplevelse is not None else bool_to_str(periode["vis_brukeropplevelse"])
    har_skjemadata_str = bool_to_str(har_skjemadata, "har_skjemadata") if har_skjemadata is not None else bool_to_str(periode["har_skjemadata"], "har_skjemadata")

    body = {
        "id": periode["id"],
        "skjema_id": skjema_id,
        "periode_type": periode["periode_type"],
        "periode_nr": periode["periode_nr"],
        "periode_aar": periode["periode_aar"],
        "periode_dato": periode_dato.isoformat()
        if periode_dato is not None
        else periode["periode_dato"],
        "delreg_nr": delreg_nr if delreg_nr is not None else periode["delreg_nr"],
        "enhet_type": enhet_type if enhet_type is not None else periode["enhet_type"],
        "vis_oppgavebyrde": vis_oppgavebyrde_str,
        "vis_brukeropplevelse": vis_brukeropplevelse_str,
        "har_skjemadata": har_skjemadata_str,
        "journalnummer": journalnummer
        if journalnummer is not None
        else periode["journalnummer"],
        "endret_av": user,
    }

    try:
        body_json = json.dumps(body)
        content: str = client.put(
            path=f"{constants.PERIODE_PATH}/{periode['id']}",
            body_json=body_json,
            context=context,
        )

        result: dict = json.loads(content)
        
        for key in ['vis_brukeropplevelse', 'vis_oppgavebyrde', 'har_skjemadata']:
            if key in result:
                result[key] = convert_to_bool(result[key])
        
        return OperationResult(value=result, log=context.logs())
    except Exception as e:
        context.set_error(
            f"Failed to update periode for skjema_id {skjema_id} - periode {periode_nr} {periode_type} {periode_nr} ",
            e,
        )
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=skjema_id_validator)
def create_periode(
    self,
    *,
    skjema_id: int,
    periode_type: str | None = None,
    periode_nr: int | None = None,
    periode_aar: int | None = None,
    periode_dato: date | None = None,
    delreg_nr: int | None = None,
    enhet_type: str | None = None,
    vis_oppgavebyrde: bool | None = False,
    vis_brukeropplevelse: bool | None = False,
    har_skjemadata: bool | None = False,
    journalnummer: str | None = None,
    context: SuvOperationContext,
) -> OperationResult:
    """
    Creates a new period with the specified details.

    Parameters:
    ------------
    skjema_id: int
        The skjema_id associated with the new period.
    periode_type: Optional[str]
        Periode type of the new periode.
    periode_aar: Optional[int]
        Year of the new periode.
    periode_nr: Optional[int]
        Periode number of the new periode.
    periode_dato: Optional[date]
        Date for the period.
    delreg_nr: Optional[int]
        delreg_nr.
    enhet_type: Optional[str]
        enhet_type.
    vis_oppgavebyrde: Optional[bool]
        A boolean flag to indicate visibility of "oppgavebyrde". Defaults to None.
    vis_brukeropplevelse: Optional[bool]
        A boolean flag to indicate visibility of "brukeropplevelse". Defaults to None.
    har_skjemadata: Optional[bool]
        A boolean flag to indicate the presence of schema data. Defaults to None.
    journalnummer: Optional[str]
        journalnummer. Defaults to None.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.

    Returns:
    --------
    OperationResult:
        An object containing the ID of the created period, or an error message if the creation fails.

    Example:
    ---------
    result = create_periode(
        skjema_id=456, periode_type="KVRT", periode_aar=2023, date(2024,12,15),periode_nr=1
    )
    """

    vis_oppgavebyrde_str = bool_to_str(vis_oppgavebyrde)
    vis_brukeropplevelse_str = bool_to_str(vis_brukeropplevelse)
    har_skjemadata_str = bool_to_str(har_skjemadata, "har_skjemadata") 

    user = user_tools.get_current_user(context)
   
   
    model = PeriodeRequestModel(
        skjema_id=skjema_id,
        endret_av=user,
        periode_type=periode_type,
        periode_nr=periode_nr,
        periode_aar=periode_aar,
        periode_dato=periode_dato.isoformat() if periode_dato else None,
        delreg_nr=delreg_nr,
        enhet_type=enhet_type,
        vis_oppgavebyrde=vis_oppgavebyrde_str,
        vis_brukeropplevelse=vis_brukeropplevelse_str,
        har_skjemadata=har_skjemadata_str,
        journalnummer=journalnummer,
    )
    
    try:
        body = model.model_dump_json()
        content: str = client.post(
            path=constants.PERIODE_PATH, body_json=body, context=context
        )
        new_id = json.loads(content)["id"]
        context.log(message="Created 'periode' with id '{new_id}'")
        
        return OperationResult(value={"id": new_id}, log=context.logs())
    except Exception as e:
        context.set_error(
            f"Failed to create for skjema_id '{skjema_id}' - periode {periode_nr} {periode_type} {periode_nr}",
            e,
        )
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=periode_id_validator)
def delete_periode(
    self, *, periode_id: int, context: SuvOperationContext
) -> OperationResult:
    """
    Deletes the period with the specified periode_id.

    Parameters:
    ------------
    periode_id: int
        The ID of the period to delete.
    context: SuvOperationContext
        Operation context for logging and error handling. This is injected by the underlying pipeline.

    Returns:
    --------
    OperationResult:
        An object containing the result of the deletion operation, or an error message if the deletion fails.

    Example:
    ---------
    result = delete_periode(
        Periode_id=123
    )
    """
    try:
        content: str = client.delete(
            path=f"{constants.PERIODE_PATH}/{periode_id}", context=context
        )
        context.log(message="Deleted 'periode' with id '{periode_id}'")
        return OperationResult(value={"result": content}, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to delete Periode with id '{periode_id}'.", e)
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )



@result_to_dict
@SuvOperationContext(validator=periode_id_validator)
def create_new_periodes(
    self,
    *,
    periode_id: int,
    num_new_periodes: int, 
    context: SuvOperationContext,
) -> OperationResult:
    """
    Creates new periodes using a given exisiting period. 

    Parameters:
    ------------
    periode_id: int
        The periode_id associated with the new periods.
    num_new_periodes: int 
        Number of new periodes to be created
   
    Returns:
    --------
    OperationResult:
        An object containing the ID of the created periodes, or an error message if the creation fails.

    Example:
    ---------
    result = create_new_periodes(periode_id=80, num_new_periodes=3)
    """

   
    
    try:
        content: str = client.post(
            path=f"{constants.PERIODE_PATH}/{periode_id}/create-new/{num_new_periodes}", body_json="", context=context
        )
        context.log(message="Created {num_new_periods} new periodes using period with id '{periode_id}'")
        return OperationResult(value={"result": content}, log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to create Periodes using period_id '{periode_id}'.", e)
        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )



import re
from datetime import date
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext


ra_pattern = re.compile("^[rR][aA]-[0-9]{4}(?:[aA]3)?$")

guid_pattern = re.compile(
    r"^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$"
)


def instance_str_validator(ctx: SuvOperationContext, kwargs):
    instance_id = kwargs.get("instance_id", None)

    instance_owner_id = int(instance_id.split("/")[0])
    instance_guid = instance_id.split("/")[1]

    if not isinstance(instance_id, str):
        raise _set_error(
            ctx,
            "Parameter 'instance_id' must be a valid string. Eg. '123451213/12345678-1234-1234-1234-123456789012'",
        )

    if not re.match(guid_pattern, instance_guid):
        raise _set_error(
            ctx,
            "Guid for instance must match pattern 'xxxxxxxx-xxxx-vxxx-xxxx-xxxxxxxxxxxx' (x = digit 0-9, a-f)",
        )

    if not isinstance(instance_owner_id, int) or instance_owner_id == -1:
        raise _set_error(ctx, "Owner for instance must be a valid positive integer.")


def skjema_id_validator(ctx: SuvOperationContext, kwargs):
    skjema_id = kwargs.get("skjema_id", -1)

    if not isinstance(skjema_id, int) or skjema_id == -1:
        raise _set_error(ctx, "Parameter 'skjema_id' must be a valid positive integer.")

def prefill_meta_id_validator(ctx: SuvOperationContext, kwargs):
    prefill_meta_id = kwargs.get("prefill_meta_id", -1)

    if not isinstance(prefill_meta_id, int) or prefill_meta_id == -1:
        raise _set_error(ctx, "Parameter 'prefill_meta_id' must be a valid positive integer.")

def ra_update_validator(ctx: SuvOperationContext, kwargs):
    ra_nummer = kwargs.get("ra_nummer", None)
    versjon = kwargs.get("versjon", -1)
    undersokelse_nr = kwargs.get("undersokelse_nr", None)

    if not isinstance(ra_nummer, str):
        raise _set_error(ctx, "Parameter 'ra_nummer' must be a valid string.")

    if not re.match(ra_pattern, ra_nummer):
        raise _set_error(
            ctx,
            "Parameter 'ra_nummer' must match pattern 'ra-XXXX' or 'RA-XXXX' (X = digit 0-9)",
        )

    if not isinstance(versjon, int) or versjon == -1:
        raise _set_error(ctx, "Parameter 'versjon' must be a valid positive integer.")

    if not isinstance(undersokelse_nr, str):
        raise _set_error(ctx, "Parameter 'undersokelse_nr' must be a valid string.")


def pulje_id_validator(ctx: SuvOperationContext, kwargs):
    pulje_id = kwargs.get("pulje_id", -1)

    if not isinstance(pulje_id, int) or pulje_id == -1:
        raise _set_error(ctx, "Parameter 'pulje_id' must be a valid positive integer.")


def utsending_id_validator(ctx: SuvOperationContext, kwargs):
    pulje_id = kwargs.get("utsending_id", -1)

    if not isinstance(pulje_id, int) or pulje_id == -1:
        raise _set_error(
            ctx, "Parameter 'utsending_id' must be a valid positive integer."
        )


def ra_nummer_validator(ctx: SuvOperationContext, kwargs):
    ra_nummer = kwargs.get("ra_nummer", None)

    if not isinstance(ra_nummer, str):
        raise _set_error(ctx, "Parameter 'ra_nummer' must be a valid positive integer.")

    if not re.match(ra_pattern, ra_nummer):
        raise _set_error(
            ctx,
            "Parameter 'ra_nummer' must match pattern 'ra-XXXX' or 'RA-XXXX' (X = digit 0-9)",
        )


def periode_id_validator(ctx: SuvOperationContext, kwargs):
    periode_id = kwargs.get("periode_id", -1)

    if not isinstance(periode_id, int) or periode_id == -1:
        raise _set_error(
            ctx, "Parameter 'periode_id' must be a valid positive integer."
        )


def delreg_id_validator(ctx: SuvOperationContext, kwargs):
    delreg_nr = kwargs.get("delreg_nr", -1)

    if not isinstance(delreg_nr, int) or delreg_nr == -1:
        raise _set_error(ctx, "Parameter 'delreg_nr' must be a valid positive integer.")


def _set_error(ctx: SuvOperationContext, message: str) -> Exception:
    ex = ValueError(message)
    ctx.set_error(error_msg=message, exception=ex)

    return ex



import json
import os
from typing import Optional

from dapla_suv_tools._internals.integration.api_client import SuvApiClient
from dapla_suv_tools._internals.util.decorators import result_to_dict
from dapla_suv_tools._internals.util.operation_result import OperationResult
from dapla_suv_tools._internals.util import constants
from dapla_suv_tools._internals.util.suv_operation_context import SuvOperationContext
from dapla_suv_tools._internals.util.validators import (
    ra_nummer_validator,
    delreg_id_validator,
)
from dapla_suv_tools.pagination import PaginationInfo


END_USER_API_BASE_URL = os.getenv("SUV_END_USER_API_URL", "")

client = SuvApiClient(base_url=END_USER_API_BASE_URL)


@result_to_dict
@SuvOperationContext(validator=ra_nummer_validator)
def get_utvalg_from_sfu(
    self,
    *,
    delreg_nr: int,
    ra_nummer: str,
    pulje: Optional[int] = 0,
    max_results: int = 0,
    pagination_info: Optional[PaginationInfo] = None,
    context: SuvOperationContext,
) -> OperationResult:
    """
    Get selection from SFU.

    Parameters:
    ------------
    delreg_nr: int, required
        The delreg number of the selection.
    ra_nummer: str, required
        Skjema's RA-number, e.g. 'RA-1234'.
    pulje: int, optional
        Limit the selection by pulje.
    context: SuvOperationContext
        Operation context.  This is injected by the underlying pipeline.  Adding a custom context will result in an error.

    Returns:
    --------
    dict:
        A list of json objects matching the selection

    Example:
    --------
    get_utvalg_from_sfu(delreg_nr="123456789", ra_nummer="123456789", pulje="123456789")

    """

    model = {
        "delreg_nr": delreg_nr,
        "ra_number": ra_nummer,
        "pulje": pulje,
    }

    try:
        content: str
        if pagination_info is None:
            content = _get_non_paged_result(
                path=f"{constants.SFU_PATH}/utvalg",
                body_json=json.dumps(model),
                max_results=max_results,
                context=context,
            )
        else:
            content = _get_paged_result(
                path=f"{constants.SFU_PATH}/utvalg",
                body_json=json.dumps(model),
                paging=pagination_info,
                context=context,
            )

        # content: str = client.post(
        #    path=f"{constants.SFU_PATH}/utvalg",
        #    body_json=json.dumps(model),
        #    context=context,
        # )
        result: dict = json.loads(content)
        context.log(message=f"Fetched selection for delreg_nr'{delreg_nr}'")

        return OperationResult(value=result["items"], log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to fetch for delreg_nr {delreg_nr}", e)

        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


@result_to_dict
@SuvOperationContext(validator=delreg_id_validator)
def get_enhet_from_sfu(
    self,
    *,
    delreg_nr: int,
    orgnr: str,
    context: SuvOperationContext,
) -> OperationResult:
    """
    Get unit from SFU.

    Parameters:
    ------------
    delreg_nr: int, required
        The delreg number of the selection.
    orgnr: str, required
        The organization number of the unit.
    context: SuvOperationContext
        Operation context.  This is injected by the underlying pipeline.  Adding a custom context will result in an error.

    Returns:
    --------
    dict:
        An object matching the organization number

    Example:
    --------
    get_enhet_from_sfu(delreg_nr="123456789", orgnr="123456789")

    """

    data = {
        "delreg_nr": delreg_nr,
        "orgnr": orgnr,
    }

    try:
        content: str = client.post(
            path=f"{constants.SFU_PATH}/enhet",
            body_json=json.dumps(data),
            context=context,
        )
        result = json.loads(content)
        context.log(message=f"Fetched org for delreg_nr'{delreg_nr}'")

        return OperationResult(value=result["items"][0], log=context.logs())
    except Exception as e:
        context.set_error(f"Failed to fetch org for delreg_nr {delreg_nr}", e)

        return OperationResult(
            success=False, value=context.errors(), log=context.logs()
        )


# Helper functions


def _get_non_paged_result(
    path: str, max_results: int, body_json: str, context: SuvOperationContext
) -> str:
    if max_results > 0:
        return client.post(
            path=f"{path}?size={max_results}&asc=false",
            body_json=body_json,
            context=context,
        )

    items = []
    total = 1
    page = 1

    while len(items) < total:
        response = client.post(
            path=f"{path}?page={page}&size=100&asc=false",
            body_json=body_json,
            context=context,
        )

        response_json = json.loads(response)
        total = int(response_json["count"])
        items.extend(response_json["items"])
        page += 1

    return json.dumps({"items": items})


def _get_paged_result(
    path: str, paging: PaginationInfo, body_json: str, context: SuvOperationContext
) -> str:
    return client.post(
        path=f"{path}?page={paging.page}&size={paging.size}&asc=false",
        body_json=body_json,
        context=context,
    )



from typing import Callable, Optional
from dapla_suv_tools._internals.util.help_helper import HelpAssistant
from dapla_suv_tools._internals.util import constants


class SuvClient:
    suppress_exceptions: bool
    operations_log: list
    help_assistant: HelpAssistant

    from dapla_suv_tools._internals.client_apis.skjema_api import (
        get_skjema_by_id,
        get_skjema_by_ra_nummer,
        create_skjema,
        delete_skjema,
        update_skjema_by_ra_number,
        update_skjema_by_id,
        get_all_skjema,
    )

    from dapla_suv_tools._internals.client_apis.periode_api import (
        get_periode_by_id,
        get_perioder_by_skjema_id,
        create_periode,
        delete_periode,
        update_periode_by_id,
        update_periode_by_skjema_id,
        create_new_periodes
    )

    from dapla_suv_tools._internals.client_apis.pulje_api import (
        get_pulje_by_id,
        get_pulje_by_periode_id,
        create_pulje,
        update_pulje_by_id,
        update_pulje_by_periode_id,
        delete_pulje,
    )
    from dapla_suv_tools._internals.client_apis.utsending_api import (
        get_utsending_by_id,
        get_utsending_by_pulje_id,
        update_utsending_by_id,
        create_utsending,
        delete_utsending,
    )

    from dapla_suv_tools._internals.client_apis.sfu_api import (
        get_utvalg_from_sfu,
        get_enhet_from_sfu,
    )

    from dapla_suv_tools._internals.client_apis.instans_api import (
        get_instance,
    )

    from dapla_suv_tools._internals.client_apis.instantiator_api import (
        resend_instances,
    )

    from dapla_suv_tools._internals.client_apis.innkvittering_api import (
        resend_receipts,
    )

    from dapla_suv_tools._internals.client_apis.prefill_meta_api import (
        get_prefill_meta_by_skjema_id,
        get_prefill_meta_by_skjema_def,
        save_prefill_meta,
        get_prefill_meta_by_prefill_meta_id,
        update_prefill_meta_by_prefill_meta_id,
        delete_skjema_prefill_meta_by_id,
        delete_skjema_prefill_meta_by_skjema_id,
    )

    from dapla_suv_tools._internals.client_apis.prefill_api_stat_api import (
        get_prefill_info_for_skjema,
        get_prefill_for_enhet,
        save_prefill_for_enhet,
        delete_prefill_for_skjema,
        delete_prefill_for_enhet,
        validate_skjemadata,
    )

    def __init__(self, suppress_exceptions: bool = False):
        self.suppress_exceptions = suppress_exceptions
        self.operations_log = []
        self._build_help_cache()

    def logs(self, threshold: str | None = None, results: int = 0) -> list:
        if not threshold or threshold not in constants.LOG_LEVELS:
            threshold = constants.LOG_INFO

        log = self._filter_logs(threshold=threshold)

        if results > 0:
            return log[-results:]

        return log

    def help(self, function: Optional[Callable] = None):
        if function is None:
            return self.__doc__
        doc = self.help_assistant.get_function_help_entry(function.__name__)

        if doc is not None:
            print(doc)

        print(f"No help entry for '{function.__name__}' exists.")

    def _build_help_cache(self):
        self.help_assistant = HelpAssistant()
        self.help_assistant.register_functions(
            [
                self.get_skjema_by_id,
                self.get_skjema_by_ra_nummer,
                self.create_skjema,
                self.delete_skjema,
                self.get_periode_by_id,
                self.get_perioder_by_skjema_id,
                self.create_periode,
                self.delete_periode,
                self.update_skjema_by_ra_number,
                self.update_skjema_by_id,
                self.get_all_skjema,
                self.update_periode_by_id,
                self.update_periode_by_skjema_id,
                self.get_pulje_by_id,
                self.get_pulje_by_periode_id,
                self.create_pulje,
                self.update_pulje_by_id,
                self.update_pulje_by_periode_id,
                self.delete_pulje,
                self.get_utsending_by_id,
                self.get_utsending_by_pulje_id,
                self.update_utsending_by_id,
                self.create_utsending,
                self.delete_utsending,
                self.get_instance,
                self.resend_instances,
                self.resend_receipts,
                self.get_prefill_meta_by_skjema_id,
                self.get_prefill_info_for_skjema,
                self.get_prefill_for_enhet,
                self.save_prefill_for_enhet,
                self.delete_prefill_for_skjema,
                self.delete_prefill_for_enhet,
            ]
        )

    def flush_logs(self):
        self.operations_log = []

    def _filter_logs(self, threshold: str) -> list:
        limit = constants.LOG_LEVELS.index(threshold)

        filtered = []

        for log_entry in self.operations_log:
            logs = log_entry["logs"]
            if len(logs) == 0:
                continue
            for entry in logs:
                if constants.LOG_LEVELS.index(entry["level"]) < limit:
                    continue
                filtered.append(entry)

        return sorted(filtered, key=lambda x: x["time"])
